@@ -14,12 +14,15 @@ import warnings
 
 if sys.hexversion < 0x03000000:
     import cPickle as pickle
+    # ISSUE #25 Fix for http://bugs.python.org/issue10211
+    from cStringIO import StringIO as BytesIO
     TextType = unicode
     BytesType = str
     INT_TYPES = int, long
     range = xrange # pylint: disable=redefined-builtin,invalid-name
 else:
     import pickle
+    from io import BytesIO # pylint: disable=ungrouped-imports
     TextType = str
     BytesType = bytes
     INT_TYPES = int,
@@ -148,7 +151,7 @@ class Disk(object):
         if raw:
             return BytesType(key) if type(key) is sqlite3.Binary else key
         else:
-            return pickle.load(io.BytesIO(key))
+            return pickle.load(BytesIO(key))
 
 
     def store(self, value, read, threshold, prep_file):
@@ -242,7 +245,7 @@ class Disk(object):
                 with io.open(op.join(directory, filename), 'rb') as reader:
                     return pickle.load(reader)
             else:
-                return pickle.load(io.BytesIO(value))
+                return pickle.load(BytesIO(value))
 
 
 class CachedAttr(object):
@@ -483,7 +486,7 @@ class Cache(with_metaclass(CacheMeta, object)):
 
 
     def set(self, key, value, expire=None, read=False, tag=None):
-        """Store key, value pair in cache.
+        """Set key, value pair in cache.
 
         When `read` is `True`, `value` should be a file-like object opened
         for reading in binary mode.
@@ -493,86 +496,116 @@ class Cache(with_metaclass(CacheMeta, object)):
         :param expire: seconds until the key expires (default None, no expiry)
         :param bool read: read value as raw bytes from file (default False)
         :param tag: text to associate with key (default None)
-        :return: True if item was successfully committed
+        :return: True if item was successfully set
 
         """
         sql = self._sql
-
+        now = time.time()
         db_key, raw = self._disk.put(key)
-
-        # Lookup filename for existing key.
-
-        rows = sql(
-            'SELECT version, filename FROM Cache WHERE key = ? AND raw = ?',
-            (db_key, raw),
-        ).fetchall()
-
-        if rows:
-            (version, filename), = rows
-        else:
-            sql('INSERT OR IGNORE INTO Cache(key, raw) VALUES (?, ?)',
-                (db_key, raw)
-            )
-            version, filename = 0, None
-
-        # Remove existing file if present.
-
-        if filename is not None:
-            self._remove(filename)
-
-        # Prepare value for disk storage.
-
+        expire_time = None if expire is None else now + expire
         size, mode, filename, db_value = self._disk.store(
             value, read, self.large_value_threshold, self._prep_file
         )
 
-        next_version = version + 1
-        now = time.time()
-        expire_time = None if expire is None else now + expire
+        try:
+            sql('BEGIN IMMEDIATE')
+        except sqlite3.OperationalError:
+            self._remove(filename)
+            raise
 
-        # Update the row. Two step process so that all files remain tracked.
+        # The order of SELECT, UPDATE, and INSERT is important below.
+        #
+        # Typical cache usage pattern is:
+        #
+        # value = cache.get(key)
+        # if value is None:
+        #     value = expensive_calculation()
+        #     cache.set(key, value)
+        #
+        # Cache.get does not evict expired keys to avoid writes during lookups.
+        # Commonly used/expired keys will therefore remain in the cache making
+        # an UPDATE the preferred path.
+        #
+        # The alternative is to assume the key is not present by first trying
+        # to INSERT and then handling the IntegrityError that occurs from
+        # violating the UNIQUE constraint. This optimistic approach was
+        # rejected based on the common cache usage pattern.
 
-        cursor = sql(
-            'UPDATE Cache SET'
-            ' version = ?,'
-            ' store_time = ?,'
-            ' expire_time = ?,'
-            ' access_time = ?,'
-            ' access_count = ?,'
-            ' tag = ?,'
-            ' size = ?,'
-            ' mode = ?,'
-            ' filename = ?,'
-            ' value = ?'
-            ' WHERE key = ? AND raw = ? AND version = ?', (
-                next_version,
-                now,          # store_time
-                expire_time,
-                now,          # access_time
-                0,            # access_count
-                tag,
-                size,
-                mode,
-                filename,
-                db_value,
-                db_key,
-                raw,
-                version,
-            ),
-        )
+        rows = sql(
+            'SELECT rowid, filename FROM Cache WHERE key = ? AND raw = ?',
+            (db_key, raw),
+        ).fetchall()
 
-        if cursor.rowcount == 0:
-            # Another Cache wrote the value before us so abort.
-            if filename is not None:
-                self._remove(filename)
-            return False
+        if rows:
+            (old_rowid, old_filename), = rows
 
-        # Evict expired keys.
+            sql('UPDATE Cache SET'
+                ' version = ?,'
+                ' store_time = ?,'
+                ' expire_time = ?,'
+                ' access_time = ?,'
+                ' access_count = ?,'
+                ' tag = ?,'
+                ' size = ?,'
+                ' mode = ?,'
+                ' filename = ?,'
+                ' value = ?'
+                ' WHERE rowid = ?', (
+                    0,            # version
+                    now,          # store_time
+                    expire_time,
+                    now,          # access_time
+                    0,            # access_count
+                    tag,
+                    size,
+                    mode,
+                    filename,
+                    db_value,
+                    old_rowid,
+                ),
+            )
+        else:
+            old_filename = None
 
+            sql('INSERT INTO Cache('
+                ' key, raw, version, store_time, expire_time, access_time,'
+                ' access_count, tag, size, mode, filename, value'
+                ') VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)', (
+                    db_key,
+                    raw,
+                    0,           # version
+                    now,         # store_time
+                    expire_time,
+                    now,         # access_time
+                    0,           # access_count
+                    tag,
+                    size,
+                    mode,
+                    filename,
+                    db_value,
+                ),
+            )
+
+        sql('COMMIT')
+
+        self._remove(old_filename)
+        self._evict_expired_keys()
+
+        return True
+
+
+    __setitem__ = set
+
+
+    def _evict_expired_keys(self):
+        "Evict expired keys."
         cull_limit = self.cull_limit
 
         if cull_limit == 0:
-            return True
+            return
+
+        sql = self._sql
+        now = time.time()
 
         rows = sql(
             'SELECT rowid, version, filename FROM Cache'
@@ -586,10 +619,10 @@ class Cache(with_metaclass(CacheMeta, object)):
             cull_limit -= deleted
 
         if cull_limit == 0:
-            return True
+            return
 
         if self.volume() < self.size_limit:
-            return True
+            return
 
         # Evict keys by policy.
 
@@ -601,22 +634,97 @@ class Cache(with_metaclass(CacheMeta, object)):
             for rowid, version, filename in rows:
                 self._delete(rowid, version, filename)
 
-        return True
 
-    __setitem__ = set
+    def add(self, key, value, expire=None, read=False, tag=None):
+        """Add key, value pair to cache.
+
+        Similar to `set`, but only set in cache if key not present.
+
+        This operation is atomic. Only one concurrent add operation for given
+        key from separate threads or processes will succeed.
+
+        When `read` is `True`, `value` should be a file-like object opened
+        for reading in binary mode.
+
+        :param key: Python key to store
+        :param value: Python value to store
+        :param float expire: seconds until the key expires
+            (default None, no expiry)
+        :param bool read: read value as bytes from file (default False)
+        :param str tag: text to associate with key (default None)
+        :return: True if item was successfully added
+
+        """
+        sql = self._sql
+        now = time.time()
+        db_key, raw = self._disk.put(key)
+        expire_time = None if expire is None else now + expire
+        size, mode, filename, db_value = self._disk.store(
+            value, read, self.large_value_threshold, self._prep_file
+        )
+
+        try:
+            sql('BEGIN IMMEDIATE')
+        except sqlite3.OperationalError:
+            self._remove(filename)
+            raise
+
+        rows = sql(
+            'SELECT rowid, filename, expire_time FROM Cache'
+            ' WHERE key = ? AND raw = ?',
+            (db_key, raw),
+        ).fetchall()
+
+        if rows:
+            (old_rowid, old_filename, old_expire_time), = rows
+
+            if old_expire_time is not None and old_expire_time < now:
+                sql('DELETE FROM Cache WHERE rowid = ?', (old_rowid,))
+            else:
+                sql('COMMIT')
+                self._remove(filename)
+                return False
+        else:
+            old_filename = None
+
+        sql('INSERT INTO Cache('
+            ' key, raw, version, store_time, expire_time, access_time,'
+            ' access_count, tag, size, mode, filename, value'
+            ') VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)', (
+                db_key,
+                raw,
+                0,           # version
+                now,         # store_time
+                expire_time,
+                now,         # access_time
+                0,           # access_count
+                tag,
+                size,
+                mode,
+                filename,
+                db_value,
+            ),
+        )
+
+        sql('COMMIT')
+
+        self._remove(old_filename)
+        self._evict_expired_keys()
+
+        return True
 
 
     def get(self, key, default=None, read=False, expire_time=False, tag=False):
-        """Get key from cache. If key is missing, return default.
+        """Retrieve value from cache. If key is missing, return default.
 
         :param key: Python key to retrieve
         :param default: value to return if key is missing (default None)
         :param bool read: if True, return file handle to value
             (default False)
-        :param float expire_time: if True, return expire_time in tuple
+        :param bool expire_time: if True, return expire_time in tuple
             (default False)
-        :param tag: if True, return tag in tuple (default False)
-        :return: key or `default` if not found
+        :param bool tag: if True, return tag in tuple (default False)
+        :return: corresponding value or `default` if not found
 
         """
         sql = self._sql
@@ -697,6 +805,20 @@ class Cache(with_metaclass(CacheMeta, object)):
         return value
 
 
+    def read(self, key):
+        """Return file handle corresponding to `key` from Cache.
+
+        :param key: Python key to retrieve
+        :return: file open for reading in binary mode
+        :raises KeyError: if key is not found
+
+        """
+        handle = self.get(key, default=ENOVAL, read=True)
+        if handle is ENOVAL:
+            raise KeyError(key)
+        return handle
+
+
     def __contains__(self, key):
         "Return True if `key` in Cache."
         sql = self._sql
@@ -759,7 +881,7 @@ class Cache(with_metaclass(CacheMeta, object)):
 
         deleted = cursor.rowcount == 1
 
-        if deleted and filename is not None:
+        if deleted:
             self._remove(filename)
 
         return deleted
@@ -784,6 +906,9 @@ class Cache(with_metaclass(CacheMeta, object)):
 
 
     def _remove(self, filename):
+        if filename is None:
+            return
+
         full_path = op.join(self._dir, filename)
 
         try:
@@ -995,6 +1120,59 @@ class Cache(with_metaclass(CacheMeta, object)):
         return count
 
 
+    def __iter__(self):
+        sql = self._sql
+        rows = sql('SELECT MAX(rowid) FROM Cache').fetchall()
+        (max_rowid,), = rows
+
+        if max_rowid is None:
+            return
+
+        chunk = self.cull_limit
+        rowid = 0
+        _disk_get = self._disk.get
+
+        while True:
+            rows = sql(
+                'SELECT rowid, key, raw FROM Cache'
+                ' WHERE ? < rowid AND rowid < ?'
+                ' ORDER BY rowid LIMIT ?',
+                (rowid, max_rowid, chunk),
+            ).fetchall()
+
+            if not rows:
+                break
+
+            for rowid, key, raw in rows:
+                yield _disk_get(key, raw)
+
+
+    def __reversed__(self):
+        sql = self._sql
+        rows = sql('SELECT MAX(rowid) FROM Cache').fetchall()
+        (rowid,), = rows
+
+        if rowid is None:
+            return
+
+        chunk = self.cull_limit
+        _disk_get = self._disk.get
+
+        while True:
+            rows = sql(
+                'SELECT rowid, key, raw FROM Cache'
+                ' WHERE 0 < rowid AND rowid < ?'
+                ' ORDER BY rowid DESC LIMIT ?',
+                (rowid, chunk),
+            ).fetchall()
+
+            if not rows:
+                break
+
+            for rowid, key, raw in rows:
+                yield _disk_get(key, raw)
+
+
     def stats(self, enable=True, reset=False):
         """Return cache statistics pair: (hits, misses).
 
